#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_mempool.h>
#include <rte_mbuf.h>
#include <rte_ether.h>
#include <rte_ip.h>
#include <rte_udp.h>
#include <rte_cycles.h>
#include <x86intrin.h>

#define NUM_PORTS 8
#define NUM_MBUFS 32768
#define MBUF_CACHE_SIZE 512
#define BURST_SIZE 256
#define PACKET_SIZE 1500
#define NUM_WORKERS 16

struct rte_mempool *mbuf_pool;
uint16_t port_id = 0;

// AVX-512 を使用したパケットの準備
void prepare_packet(struct rte_mbuf *pkt) {
    __m512i ether_hdr = _mm512_set_epi64(  // Ethernetヘッダー
        0x020000000000, 0x010000000000, 0x08004500, 0x00003C0000, 
        0x4006117B, 0xC0A80101, 0xC0A80102, 0x00000000
    );
    
    __m512i udp_hdr = _mm512_set_epi64(    // UDPヘッダー
        0x30390000, 0x003C0000, 0x00000000, 0x00000000,
        0x00000000, 0x00000000, 0x00000000, 0x00000000
    );

    // Mbuf に直接書き込む
    uint8_t *pkt_data = rte_pktmbuf_mtod(pkt, uint8_t *);
    _mm512_storeu_si512((__m512i*)pkt_data, ether_hdr);
    _mm512_storeu_si512((__m512i*)(pkt_data + 64), udp_hdr);

    pkt->data_len = PACKET_SIZE;
    pkt->pkt_len = PACKET_SIZE;
}

// NICへパケット送信 (AVX-512 を使ったバースト最適化)
void send_packets() {
    struct rte_mbuf *pkts[BURST_SIZE];
    
    for (int i = 0; i < BURST_SIZE; i++) {
        pkts[i] = rte_pktmbuf_alloc(mbuf_pool);
        prepare_packet(pkts[i]);
    }

    while (1) {
        uint16_t sent = 0;

        // アセンブリ最適化 (AVX-512)
        __asm__ __volatile__ (
            "vmovdqa64 %%zmm0, (%0)\n\t"
            "vmovdqa64 %%zmm1, (%1)\n\t"
            :
            : "r"(pkts), "r"(port_id)
            : "zmm0", "zmm1"
        );

        // バースト送信
        sent = rte_eth_tx_burst(port_id, 0, pkts, BURST_SIZE);
        if (sent < BURST_SIZE) {
            for (uint16_t i = sent; i < BURST_SIZE; i++) {
                rte_pktmbuf_free(pkts[i]);
            }
        }
    }
}

int main(int argc, char *argv[]) {
    int ret = rte_eal_init(argc, argv);
    if (ret < 0) rte_exit(EXIT_FAILURE, "EAL初期化失敗\n");

    mbuf_pool = rte_mempool_create("MBUF_POOL", NUM_MBUFS, 
                RTE_MBUF_DEFAULT_BUF_SIZE, MBUF_CACHE_SIZE, sizeof(struct rte_pktmbuf_pool_private),
                rte_socket_id(), NULL);
    if (mbuf_pool == NULL) rte_exit(EXIT_FAILURE, "メモリプール作成失敗\n");

    ret = rte_eth_dev_configure(port_id, 1, 1, NULL);
    if (ret < 0) rte_exit(EXIT_FAILURE, "ポート設定失敗\n");

    ret = rte_eth_tx_queue_setup(port_id, 0, 128, rte_socket_id(), NULL);
    if (ret < 0) rte_exit(EXIT_FAILURE, "TXキュー設定失敗\n");

    ret = rte_eth_dev_start(port_id);
    if (ret < 0) rte_exit(EXIT_FAILURE, "ポート起動失敗\n");

    printf("DPDK 初期化完了: ポート %d で送信開始\n", port_id);

    send_packets();

    return 0;
}
